<?php

namespace Drupal\cache_tags_simplify\EventSubscriber;

use Drupal\Core\Cache\CacheableResponseInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\HttpKernel\Event\ResponseEvent;
use Symfony\Component\HttpKernel\KernelEvents;

/**
 * Simplify the list of cache tags that will be put in the HTTP headers.
 *
 * Simplification aims at reducing redundant cache tags from the list of
 * tags. For example we have a view that shows 5 nodes and adds all of their
 * tags and also node_list tag. The result here should drop all concrete
 * `node:id` tags, as they are already covered by the main list one.
 *
 * Note that this is not limited for nodes or list tags only. Custom code can
 * consider other redundant scenarios and add them through implementing hook
 * `hook_cache_tags_simplify_dictionary` implementation.
 *
 * @see hook_cache_tags_simplify_dictionary()
 */
class CacheTagsSimplifySubscriber implements EventSubscriberInterface {

  /**
   * Static dictionary to use for simplification.
   *
   * @var string[]
   */
  protected $dictionary = [];

  /**
   * Class constructor.
   */
  public function __construct(ModuleHandlerInterface $module_handler) {
    $this->dictionary = $module_handler->invokeAll('cache_tags_simplify_dictionary');
    $module_handler->alter('cache_tags_simplify_dictionary', $this->dictionary);
  }

  /**
   * {@inheritDoc}
   *
   * @see \Drupal\purge\EventSubscriber\CacheableResponseSubscriber::getSubscribedEvents()
   */
  public static function getSubscribedEvents(): array {
    return [
      // To run before Purge's CacheableResponseSubscriber.
      KernelEvents::RESPONSE => ['onRespond', 1],
    ];
  }

  /**
   * Event "onRespond" handling.
   */
  public function onRespond(ResponseEvent $event): void {
    $response = $event->getResponse();

    if ($response instanceof CacheableResponseInterface) {
      $cache_tags = $response->getCacheableMetadata()->getCacheTags();

      // For each list cache tag, we get the matching cache tags. It will be
      // useful later.
      $matches = [];

      foreach ($this->dictionary as $list_tag => $pattern) {
        $matches[$list_tag] = [];

        foreach ($cache_tags as $tag) {
          if (preg_match($pattern, $tag)) {
            $matches[$list_tag][] = $tag;
          }
        }
      }

      // Cache tags simplification: we say the same with fewer words. Always
      // applies as it's harmless.
      foreach ($matches as $list_tag => $matching_tags) {
        if (\in_array($list_tag, $cache_tags, TRUE)) {
          $cache_tags = array_diff($cache_tags, $matching_tags);
        }
      }

      // Cache tags replacement: we replace many cache tags with another one
      // which will always invalidate on the same events but also on other
      // events.
      // This can lead to useless invalidations, decreasing cache hit ratio.
      // So we apply it only if it's needed to get under the cache tags count
      // limit.
      if (
        ($limit = \Drupal::config('cache_tags_simplify.settings')
          ->get('max_cache_tags_count'))
        && \count($cache_tags) > $limit
      ) {
        // We count how many tags we could replace for each possible replacement
        // and sort by this count (it's better to replace 10000 node:N with 1
        // node_list than 2 taxonomy_term:N with 1 taxonomy_term_list).
        $counts = [];

        foreach ($matches as $list_tag => $matching_tags) {
          $counts[$list_tag] = \count($matching_tags);
        }
        arsort($counts);
        $candidates = array_keys(array_filter($counts, static function ($count) {
          // It's useless to replace 1 or 0 tag with 1 list tag.
          return $count > 1;
        }));

        // Replace until we get under the limit.
        if (!empty($candidates)) {
          do {
            $list_tag = array_shift($candidates);
            $cache_tags[] = $list_tag;
            $cache_tags = array_diff($cache_tags, $matches[$list_tag]);
          } while (!empty($candidates) && \count($cache_tags) > $limit);
        }

        if (\count($cache_tags) > $limit) {
          // We still have too many cache tags but we don't know how to have
          // less.
          // So we remove all cache tags and set the response to uncacheable.
          $response->getCacheableMetadata()->setCacheTags([]);
          $response->getCacheableMetadata()->setCacheMaxAge(0);

          return;
        }
      }

      $response->getCacheableMetadata()->setCacheTags($cache_tags);
    }
  }

}
